@@ -1,10 +1,11 @@
 use std::convert::TryFrom;
 
-use afarray::Array;
+use afarray::{Array, ArrayExt, Coords};
 use async_trait::async_trait;
 use futures::future::{self, TryFutureExt};
 use futures::stream::{self, StreamExt, TryStreamExt};
 use futures::try_join;
+use log::debug;
 
 use tc_btree::*;
 use tc_error::*;
@@ -15,13 +16,17 @@ use tcgeneric::{TCBoxTryFuture, TCStream, TCTryStream};
 
 use crate::sparse::{SparseAccess, SparseAccessor};
 use crate::stream::{Read, ReadValueAt};
-use crate::transform::{self, Rebase};
+use crate::transform;
 use crate::{Bounds, Coord, Phantom, Shape, TensorAccess, TensorType, ERR_NONBIJECTIVE_WRITE};
 
 use super::file::{BlockListFile, BlockListFileSlice};
 use super::stream::SparseValueStream;
 use super::{DenseTensor, PER_BLOCK};
 
+const ERR_NON_SEQUENTIAL_WRITE: &str =
+    "non-sequential writes to a dense Tensor are not efficient--\
+consider transposing or broadcasting the value instead";
+
 /// Common [`DenseTensor`] access methods
 #[async_trait]
 pub trait DenseAccess<FD: File<Array>, FS: File<Node>, D: Dir, T: Transaction<D>>:
@@ -57,12 +62,7 @@ pub trait DenseAccess<FD: File<Array>, FS: File<Node>, D: Dir, T: Transaction<D>
 
             let values = values
                 .map_ok(|array| array.to_vec())
-                .map_ok(|values| {
-                    values
-                        .into_iter()
-                        .map(Ok)
-                        .collect::<Vec<TCResult<Number>>>()
-                })
+                .map_ok(|values| values.into_iter().map(Ok))
                 .map_ok(futures::stream::iter)
                 .try_flatten();
 
@@ -77,11 +77,14 @@ pub trait DenseAccess<FD: File<Array>, FS: File<Node>, D: Dir, T: Transaction<D>
     /// Return a transpose of this [`DenseTensor`].
     fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose>;
 
+    /// Return an Array with the values at the given coordinates.
+    async fn read_values(self, txn: Self::Txn, coords: Coords) -> TCResult<Array>;
+
+    /// Overwrite this accessor's contents with those of the given accessor.
+    async fn write<V: DenseAccess<FD, FS, D, T>>(&self, txn: Self::Txn, value: V) -> TCResult<()>;
+
     /// Write a value to the slice of this [`DenseTensor`] with the given [`Bounds`].
     async fn write_value(&self, txn_id: TxnId, bounds: Bounds, number: Number) -> TCResult<()>;
-
-    /// Overwrite a single element of this [`DenseTensor`].
-    fn write_value_at(&self, txn_id: TxnId, coord: Coord, value: Number) -> TCBoxTryFuture<()>;
 }
 
 /// A generic enum which can contain any [`DenseAccess`] impl
@@ -264,6 +267,36 @@ where
         }
     }
 
+    async fn read_values(self, txn: Self::Txn, coords: Coords) -> TCResult<Array> {
+        match self {
+            Self::File(file) => file.read_values(txn, coords).await,
+            Self::Slice(slice) => slice.read_values(txn, coords).await,
+            Self::Broadcast(broadcast) => broadcast.read_values(txn, coords).await,
+            Self::Cast(cast) => cast.read_values(txn, coords).await,
+            Self::Combine(combine) => combine.read_values(txn, coords).await,
+            Self::Expand(expansion) => expansion.read_values(txn, coords).await,
+            Self::Reduce(reduced) => reduced.read_values(txn, coords).await,
+            Self::Sparse(sparse) => sparse.read_values(txn, coords).await,
+            Self::Transpose(transpose) => transpose.read_values(txn, coords).await,
+            Self::Unary(unary) => unary.read_values(txn, coords).await,
+        }
+    }
+
+    async fn write<V: DenseAccess<FD, FS, D, T>>(&self, txn: Self::Txn, value: V) -> TCResult<()> {
+        match self {
+            Self::File(file) => file.write(txn, value).await,
+            Self::Slice(slice) => slice.write(txn, value).await,
+            Self::Broadcast(broadcast) => broadcast.write(txn, value).await,
+            Self::Cast(cast) => cast.write(txn, value).await,
+            Self::Combine(combine) => combine.write(txn, value).await,
+            Self::Expand(expansion) => expansion.write(txn, value).await,
+            Self::Reduce(reduced) => reduced.write(txn, value).await,
+            Self::Sparse(sparse) => sparse.write(txn, value).await,
+            Self::Transpose(transpose) => transpose.write(txn, value).await,
+            Self::Unary(unary) => unary.write(txn, value).await,
+        }
+    }
+
     async fn write_value(&self, txn_id: TxnId, bounds: Bounds, number: Number) -> TCResult<()> {
         match self {
             Self::File(file) => file.write_value(txn_id, bounds, number).await,
@@ -278,21 +311,6 @@ where
             Self::Unary(unary) => unary.write_value(txn_id, bounds, number).await,
         }
     }
-
-    fn write_value_at(&self, txn_id: TxnId, coord: Coord, value: Number) -> TCBoxTryFuture<()> {
-        match self {
-            Self::File(file) => file.write_value_at(txn_id, coord, value),
-            Self::Slice(slice) => slice.write_value_at(txn_id, coord, value),
-            Self::Broadcast(broadcast) => broadcast.write_value_at(txn_id, coord, value),
-            Self::Cast(cast) => cast.write_value_at(txn_id, coord, value),
-            Self::Combine(combine) => combine.write_value_at(txn_id, coord, value),
-            Self::Expand(expansion) => expansion.write_value_at(txn_id, coord, value),
-            Self::Reduce(reduced) => reduced.write_value_at(txn_id, coord, value),
-            Self::Sparse(sparse) => sparse.write_value_at(txn_id, coord, value),
-            Self::Transpose(transpose) => transpose.write_value_at(txn_id, coord, value),
-            Self::Unary(unary) => unary.write_value_at(txn_id, coord, value),
-        }
-    }
 }
 
 impl<FD, FS, D, T> ReadValueAt<D> for DenseAccessor<FD, FS, D, T>
@@ -361,7 +379,7 @@ where
     ) -> TCResult<Self> {
         if left.shape() != right.shape() {
             return Err(TCError::bad_request(
-                format!("Cannot combine shape {} with shape", left.shape()),
+                format!("cannot combine shape {} with shape", left.shape()),
                 right.shape(),
             ));
         }
@@ -433,6 +451,8 @@ where
     }
 
     fn block_stream<'a>(self, txn: T) -> TCBoxTryFuture<'a, TCTryStream<'a, Array>> {
+        debug!("BlockListCombine::block_stream");
+
         Box::pin(async move {
             let left = self.left.block_stream(txn.clone());
             let right = self.right.block_stream(txn);
@@ -442,9 +462,7 @@ where
             let blocks = left
                 .zip(right)
                 .map(|(l, r)| Ok((l?, r?)))
-                .map_ok(move |(l, r)| {
-                    combinator(&l, &r)
-                });
+                .map_ok(move |(l, r)| combinator(&l, &r));
 
             let blocks: TCTryStream<'a, Array> = Box::pin(blocks);
             Ok(blocks)
@@ -477,14 +495,25 @@ where
         )
     }
 
-    async fn write_value(&self, _txn_id: TxnId, _bounds: Bounds, _number: Number) -> TCResult<()> {
+    async fn read_values(self, txn: Self::Txn, coords: Coords) -> TCResult<Array> {
+        let (left, right) = try_join!(
+            self.left.read_values(txn.clone(), coords.clone()),
+            self.right.read_values(txn, coords)
+        )?;
+
+        Ok((self.combinator)(&left, &right))
+    }
+
+    async fn write<V: DenseAccess<FD, FS, D, T>>(
+        &self,
+        _txn: Self::Txn,
+        _value: V,
+    ) -> TCResult<()> {
         Err(TCError::unsupported(ERR_NONBIJECTIVE_WRITE))
     }
 
-    fn write_value_at(&self, _txn_id: TxnId, _coord: Coord, _value: Number) -> TCBoxTryFuture<()> {
-        Box::pin(future::ready(Err(TCError::unsupported(
-            ERR_NONBIJECTIVE_WRITE,
-        ))))
+    async fn write_value(&self, _txn_id: TxnId, _bounds: Bounds, _number: Number) -> TCResult<()> {
+        Err(TCError::unsupported(ERR_NONBIJECTIVE_WRITE))
     }
 }
 
@@ -557,7 +586,7 @@ where
     }
 
     fn size(&self) -> u64 {
-        self.source.size()
+        self.shape().size()
     }
 }
 
@@ -585,23 +614,34 @@ where
         DenseAccessor::Broadcast(Box::new(broadcast))
     }
 
-    // TODO: replace with block_stream
-    fn value_stream<'a>(self, txn: T) -> TCBoxTryFuture<'a, TCTryStream<'a, Number>> {
-        let bounds = Bounds::all(self.shape());
-        let values = stream::iter(bounds.affected()).then(move |coord| {
-            self.clone()
-                .read_value_at(txn.clone(), coord)
-                .map_ok(|(_, value)| value)
-        });
+    fn block_stream<'a>(self, txn: T) -> TCBoxTryFuture<'a, TCTryStream<'a, Array>> {
+        let shape = self.shape().clone();
+        let size = self.size();
+        let rebase = self.rebase;
+        let source = self.source;
 
-        let values: TCTryStream<'a, Number> = Box::pin(values);
-        Box::pin(future::ready(Ok(values)))
+        let blocks = stream::iter((0..size).step_by(PER_BLOCK))
+            .map(move |start| {
+                let end = match start + PER_BLOCK as u64 {
+                    end if end > size => size,
+                    end => end,
+                };
+
+                ArrayExt::range(start, end)
+            })
+            .map(move |offsets| Coords::from_offsets(offsets, &shape))
+            .map(move |coords| rebase.invert_coords(&coords))
+            .map(move |coords| source.clone().read_values(txn.clone(), coords))
+            .buffered(num_cpus::get());
+
+        let blocks: TCTryStream<'a, Array> = Box::pin(blocks);
+        Box::pin(future::ready(Ok(blocks)))
     }
 
     fn slice(self, bounds: Bounds) -> TCResult<Self::Slice> {
         self.shape().validate_bounds(&bounds)?;
 
-        let shape = bounds.to_shape();
+        let shape = bounds.to_shape(self.shape())?;
         let bounds = self.rebase.invert_bounds(bounds);
         let source = self.source.slice(bounds)?;
         BlockListBroadcast::new(source, shape)
@@ -611,14 +651,21 @@ where
         BlockListTranspose::new(self, permutation)
     }
 
-    async fn write_value(&self, _txn_id: TxnId, _bounds: Bounds, _number: Number) -> TCResult<()> {
+    async fn read_values(self, txn: Self::Txn, coords: Coords) -> TCResult<Array> {
+        let coords = self.rebase.invert_coords(&coords);
+        self.source.read_values(txn, coords).await
+    }
+
+    async fn write<V: DenseAccess<FD, FS, D, T>>(
+        &self,
+        _txn: Self::Txn,
+        _value: V,
+    ) -> TCResult<()> {
         Err(TCError::unsupported(ERR_NONBIJECTIVE_WRITE))
     }
 
-    fn write_value_at(&self, _txn_id: TxnId, _coord: Coord, _value: Number) -> TCBoxTryFuture<()> {
-        Box::pin(future::ready(Err(TCError::unsupported(
-            ERR_NONBIJECTIVE_WRITE,
-        ))))
+    async fn write_value(&self, _txn_id: TxnId, _bounds: Bounds, _number: Number) -> TCResult<()> {
+        Err(TCError::unsupported(ERR_NONBIJECTIVE_WRITE))
     }
 }
 
@@ -731,12 +778,21 @@ where
         Ok(BlockListCast::new(transpose, self.dtype))
     }
 
-    async fn write_value(&self, txn_id: TxnId, bounds: Bounds, number: Number) -> TCResult<()> {
-        self.source.write_value(txn_id, bounds, number).await
+    async fn read_values(self, txn: Self::Txn, coords: Coords) -> TCResult<Array> {
+        let dtype = self.dtype;
+
+        self.source
+            .read_values(txn, coords)
+            .map_ok(|values| values.cast_into(dtype))
+            .await
     }
 
-    fn write_value_at(&self, txn_id: TxnId, coord: Coord, value: Number) -> TCBoxTryFuture<()> {
-        self.source.write_value_at(txn_id, coord, value)
+    async fn write<V: DenseAccess<FD, FS, D, T>>(&self, txn: Self::Txn, value: V) -> TCResult<()> {
+        self.source.write(txn, value).await
+    }
+
+    async fn write_value(&self, txn_id: TxnId, bounds: Bounds, number: Number) -> TCResult<()> {
+        self.source.write_value(txn_id, bounds, number).await
     }
 }
 
@@ -853,15 +909,19 @@ where
         self.source.transpose(permutation) // TODO: expand the result
     }
 
+    async fn read_values(self, txn: Self::Txn, coords: Coords) -> TCResult<Array> {
+        let coords = self.rebase.invert_coords(&coords);
+        self.source.read_values(txn, coords).await
+    }
+
+    async fn write<V: DenseAccess<FD, FS, D, T>>(&self, txn: Self::Txn, value: V) -> TCResult<()> {
+        self.source.write(txn, value).await
+    }
+
     async fn write_value(&self, txn_id: TxnId, bounds: Bounds, number: Number) -> TCResult<()> {
         let bounds = self.rebase.invert_bounds(bounds);
         self.source.write_value(txn_id, bounds, number).await
     }
-
-    fn write_value_at(&self, txn_id: TxnId, coord: Coord, value: Number) -> TCBoxTryFuture<()> {
-        let coord = self.rebase.invert_coord(&coord);
-        self.source.write_value_at(txn_id, coord, value)
-    }
 }
 
 impl<FD, FS, D, T, B> ReadValueAt<D> for BlockListExpand<FD, FS, D, T, B>
@@ -966,16 +1026,18 @@ where
 
     fn value_stream<'a>(self, txn: T) -> TCBoxTryFuture<'a, TCTryStream<'a, Number>> {
         Box::pin(async move {
-            let values = stream::iter(Bounds::all(self.shape()).affected()).then(move |coord| {
-                let txn = txn.clone();
-                let source = self.source.clone();
-                let reductor = self.reductor;
-                let source_bounds = self.rebase.invert_coord(&coord);
-                Box::pin(async move {
-                    let slice = source.slice(source_bounds)?;
-                    reductor(&slice.accessor().into(), txn.clone()).await
+            let values = stream::iter(Bounds::all(self.shape()).affected())
+                .map(move |coord| {
+                    let txn = txn.clone();
+                    let source = self.source.clone();
+                    let reductor = self.reductor;
+                    let source_bounds = self.rebase.invert_coord(&coord);
+                    Box::pin(async move {
+                        let slice = source.slice(source_bounds)?;
+                        reductor(&slice.accessor().into(), txn.clone()).await
+                    })
                 })
-            });
+                .buffered(num_cpus::get());
 
             let values: TCTryStream<'a, Number> = Box::pin(values);
             Ok(values)
@@ -997,14 +1059,28 @@ where
         BlockListReduce::new(source, reduce_axis, self.reductor)
     }
 
-    async fn write_value(&self, _txn_id: TxnId, _bounds: Bounds, _number: Number) -> TCResult<()> {
+    async fn read_values(self, txn: Self::Txn, coords: Coords) -> TCResult<Array> {
+        let coords = coords.into_vec();
+        let values: Vec<Number> = stream::iter(coords)
+            .map(move |coord| self.clone().read_value_at(txn.clone(), coord))
+            .buffered(num_cpus::get())
+            .map_ok(|(_coord, value)| value)
+            .try_collect()
+            .await?;
+
+        Ok(Array::from(values))
+    }
+
+    async fn write<V: DenseAccess<FD, FS, D, T>>(
+        &self,
+        _txn: Self::Txn,
+        _value: V,
+    ) -> TCResult<()> {
         Err(TCError::unsupported(ERR_NONBIJECTIVE_WRITE))
     }
 
-    fn write_value_at(&self, _txn_id: TxnId, _coord: Coord, _value: Number) -> TCBoxTryFuture<()> {
-        Box::pin(future::ready(Err(TCError::unsupported(
-            ERR_NONBIJECTIVE_WRITE,
-        ))))
+    async fn write_value(&self, _txn_id: TxnId, _bounds: Bounds, _number: Number) -> TCResult<()> {
+        Err(TCError::unsupported(ERR_NONBIJECTIVE_WRITE))
     }
 }
 
@@ -1118,15 +1194,22 @@ where
         Err(TCError::not_implemented("BlockListTranspose::transpose"))
     }
 
+    async fn read_values(self, _txn: Self::Txn, _coords: Coords) -> TCResult<Array> {
+        Err(TCError::not_implemented("BlockListTranspose::read_values"))
+    }
+
+    async fn write<V: DenseAccess<FD, FS, D, T>>(
+        &self,
+        _txn: Self::Txn,
+        _value: V,
+    ) -> TCResult<()> {
+        Err(TCError::unsupported(ERR_NON_SEQUENTIAL_WRITE))
+    }
+
     async fn write_value(&self, txn_id: TxnId, bounds: Bounds, number: Number) -> TCResult<()> {
         let bounds = self.rebase.invert_bounds(bounds);
         self.source.write_value(txn_id, bounds, number).await
     }
-
-    fn write_value_at(&self, txn_id: TxnId, coord: Coord, value: Number) -> TCBoxTryFuture<()> {
-        let coord = self.rebase.invert_coord(&coord);
-        self.source.write_value_at(txn_id, coord, value)
-    }
 }
 
 impl<FD, FS, D, T, B> ReadValueAt<D> for BlockListTranspose<FD, FS, D, T, B>
@@ -1204,6 +1287,8 @@ where
     }
 
     fn value_stream<'a>(self, txn: T) -> TCBoxTryFuture<'a, TCTryStream<'a, Number>> {
+        debug!("BlockListSparse::value_stream");
+
         Box::pin(async move {
             let bounds = Bounds::all(self.shape());
             let zero = self.dtype().zero();
@@ -1224,6 +1309,41 @@ where
         Ok(transpose.into())
     }
 
+    async fn read_values(self, txn: Self::Txn, coords: Coords) -> TCResult<Array> {
+        let coords = coords.into_vec();
+        let source = self.source.clone();
+        let values: Vec<Number> = stream::iter(coords)
+            .map(move |coord| source.clone().read_value_at(txn.clone(), coord))
+            .buffered(num_cpus::get())
+            .map_ok(|(_coord, value)| value)
+            .try_collect()
+            .await?;
+
+        Ok(Array::from(values))
+    }
+
+    async fn write<V: DenseAccess<FD, FS, D, T>>(&self, txn: Self::Txn, value: V) -> TCResult<()> {
+        if value.shape() != self.shape() {
+            return Err(TCError::unsupported(format!(
+                "cannot write a value of shape {} to a tensor of shape {}",
+                value.shape(),
+                self.shape()
+            )));
+        }
+
+        let txn_id = *txn.id();
+        let bounds = Bounds::all(self.shape());
+        let values = value.value_stream(txn).await?;
+        let source = &self.source;
+        stream::iter(bounds.affected())
+            .zip(values)
+            .map(|(coord, r)| r.map(|n| (coord, n)))
+            .map_ok(|(coord, n)| source.write_value(txn_id, coord, n))
+            .try_buffer_unordered(num_cpus::get())
+            .try_fold((), |_, _| future::ready(Ok(())))
+            .await
+    }
+
     async fn write_value(&self, txn_id: TxnId, bounds: Bounds, number: Number) -> TCResult<()> {
         stream::iter(bounds.affected())
             .map(|coord| self.source.write_value(txn_id, coord, number))
@@ -1231,10 +1351,6 @@ where
             .try_fold((), |_, _| future::ready(Ok(())))
             .await
     }
-
-    fn write_value_at(&self, txn_id: TxnId, coord: Coord, value: Number) -> TCBoxTryFuture<()> {
-        self.source.write_value(txn_id, coord, value)
-    }
 }
 
 impl<FD, FS, D, T, A> ReadValueAt<D> for BlockListSparse<FD, FS, D, T, A>
@@ -1377,14 +1493,25 @@ where
         })
     }
 
-    async fn write_value(&self, _txn_id: TxnId, _bounds: Bounds, _number: Number) -> TCResult<()> {
+    async fn read_values(self, txn: Self::Txn, coords: Coords) -> TCResult<Array> {
+        let transform = self.transform;
+
+        self.source
+            .read_values(txn, coords)
+            .map_ok(move |values| (transform)(&values))
+            .await
+    }
+
+    async fn write<V: DenseAccess<FD, FS, D, T>>(
+        &self,
+        _txn: Self::Txn,
+        _value: V,
+    ) -> TCResult<()> {
         Err(TCError::unsupported(ERR_NONBIJECTIVE_WRITE))
     }
 
-    fn write_value_at(&self, _txn_id: TxnId, _coord: Coord, _value: Number) -> TCBoxTryFuture<()> {
-        Box::pin(future::ready(Err(TCError::unsupported(
-            ERR_NONBIJECTIVE_WRITE,
-        ))))
+    async fn write_value(&self, _txn_id: TxnId, _bounds: Bounds, _number: Number) -> TCResult<()> {
+        Err(TCError::unsupported(ERR_NONBIJECTIVE_WRITE))
     }
 }
 
