@@ -8,7 +8,7 @@ use arrayfire as af;
 use async_trait::async_trait;
 use destream::{de, en, EncodeSeq};
 use futures::future::{self, TryFutureExt};
-use futures::stream::{Stream, StreamExt, TryStreamExt};
+use futures::stream::{Stream, TryStreamExt};
 use log::debug;
 
 use tc_btree::Node;
@@ -29,6 +29,7 @@ use super::{
 use access::*;
 pub use access::{BlockListSparse, DenseAccess, DenseAccessor};
 pub use file::BlockListFile;
+use futures::StreamExt;
 
 mod access;
 mod file;
@@ -130,8 +131,8 @@ where
 
 impl<FD, FS, D, T, B> TensorAccess for DenseTensor<FD, FS, D, T, B>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
+    FD: File<Array>,
+    FS: File<Node>,
     D: Dir,
     T: Transaction<D>,
     B: DenseAccess<FD, FS, D, T>,
@@ -225,8 +226,7 @@ where
     }
 }
 
-#[async_trait]
-impl<FD, FS, D, T, B, O> TensorCompare<D, DenseTensor<FD, FS, D, T, O>>
+impl<FD, FS, D, T, B, O> TensorCompare<DenseTensor<FD, FS, D, T, O>>
     for DenseTensor<FD, FS, D, T, B>
 where
     FD: File<Array> + TryFrom<D::File, Error = TCError>,
@@ -237,15 +237,10 @@ where
     O: DenseAccess<FD, FS, D, T>,
     D::FileClass: From<TensorType>,
 {
-    type Txn = T;
     type Compare = DenseTensor<FD, FS, D, T, BlockListCombine<FD, FS, D, T, B, O>>;
     type Dense = DenseTensor<FD, FS, D, T, BlockListCombine<FD, FS, D, T, B, O>>;
 
-    async fn eq(
-        self,
-        other: DenseTensor<FD, FS, D, T, O>,
-        _txn: Self::Txn,
-    ) -> TCResult<Self::Dense> {
+    fn eq(self, other: DenseTensor<FD, FS, D, T, O>) -> TCResult<Self::Dense> {
         fn eq(l: Number, r: Number) -> Number {
             Number::from(l == r)
         }
@@ -261,11 +256,7 @@ where
         self.combine(other, Array::gt, gt, NumberType::Bool)
     }
 
-    async fn gte(
-        self,
-        other: DenseTensor<FD, FS, D, T, O>,
-        _txn: Self::Txn,
-    ) -> TCResult<Self::Dense> {
+    fn gte(self, other: DenseTensor<FD, FS, D, T, O>) -> TCResult<Self::Dense> {
         fn gte(l: Number, r: Number) -> Number {
             Number::from(l >= r)
         }
@@ -281,11 +272,7 @@ where
         self.combine(other, Array::lt, lt, NumberType::Bool)
     }
 
-    async fn lte(
-        self,
-        other: DenseTensor<FD, FS, D, T, O>,
-        _txn: Self::Txn,
-    ) -> TCResult<Self::Dense> {
+    fn lte(self, other: DenseTensor<FD, FS, D, T, O>) -> TCResult<Self::Dense> {
         fn lte(l: Number, r: Number) -> Number {
             Number::from(l > r)
         }
@@ -302,8 +289,7 @@ where
     }
 }
 
-#[async_trait]
-impl<FD, FS, D, T, B> TensorCompare<D, Tensor<FD, FS, D, T>> for DenseTensor<FD, FS, D, T, B>
+impl<FD, FS, D, T, B> TensorCompare<Tensor<FD, FS, D, T>> for DenseTensor<FD, FS, D, T, B>
 where
     FD: File<Array> + TryFrom<D::File, Error = TCError>,
     FS: File<Node> + TryFrom<D::File, Error = TCError>,
@@ -312,19 +298,13 @@ where
     B: DenseAccess<FD, FS, D, T>,
     D::FileClass: From<TensorType>,
 {
-    type Txn = T;
     type Compare = Tensor<FD, FS, D, T>;
     type Dense = Tensor<FD, FS, D, T>;
 
-    async fn eq(self, other: Tensor<FD, FS, D, T>, txn: Self::Txn) -> TCResult<Self::Dense> {
+    fn eq(self, other: Tensor<FD, FS, D, T>) -> TCResult<Self::Dense> {
         match other {
-            Tensor::Dense(dense) => self.eq(dense, txn).map_ok(Tensor::from).await,
-            Tensor::Sparse(sparse) => {
-                self.into_sparse()
-                    .eq(sparse, txn)
-                    .map_ok(Tensor::from)
-                    .await
-            }
+            Tensor::Dense(dense) => self.eq(dense).map(Tensor::from),
+            Tensor::Sparse(sparse) => self.into_sparse().eq(sparse).map(Tensor::from),
         }
     }
 
@@ -335,14 +315,10 @@ where
         }
     }
 
-    async fn gte(self, other: Tensor<FD, FS, D, T>, txn: Self::Txn) -> TCResult<Self::Dense> {
+    fn gte(self, other: Tensor<FD, FS, D, T>) -> TCResult<Self::Dense> {
         match other {
-            Tensor::Dense(dense) => self.gte(dense, txn).map_ok(Tensor::from).await,
-            Tensor::Sparse(sparse) => {
-                self.gte(sparse.into_dense(), txn)
-                    .map_ok(Tensor::from)
-                    .await
-            }
+            Tensor::Dense(dense) => self.gte(dense).map(Tensor::from),
+            Tensor::Sparse(sparse) => self.gte(sparse.into_dense()).map(Tensor::from),
         }
     }
 
@@ -353,14 +329,10 @@ where
         }
     }
 
-    async fn lte(self, other: Tensor<FD, FS, D, T>, txn: Self::Txn) -> TCResult<Self::Dense> {
+    fn lte(self, other: Tensor<FD, FS, D, T>) -> TCResult<Self::Dense> {
         match other {
-            Tensor::Dense(dense) => self.lte(dense, txn).map_ok(Tensor::from).await,
-            Tensor::Sparse(sparse) => {
-                self.lte(sparse.into_dense(), txn)
-                    .map_ok(Tensor::from)
-                    .await
-            }
+            Tensor::Dense(dense) => self.lte(dense).map(Tensor::from),
+            Tensor::Sparse(sparse) => self.lte(sparse.into_dense()).map(Tensor::from),
         }
     }
 
@@ -392,11 +364,15 @@ where
     }
 
     async fn write_value(&self, txn_id: TxnId, bounds: Bounds, value: Number) -> TCResult<()> {
-        self.blocks.clone().write_value(txn_id, bounds, value).await
+        self.blocks.write_value(txn_id, bounds, value).await
     }
 
     async fn write_value_at(&self, txn_id: TxnId, coord: Coord, value: Number) -> TCResult<()> {
-        self.blocks.write_value_at(txn_id, coord, value).await
+        debug!("DenseTensor::write_value_at");
+
+        self.blocks
+            .write_value(txn_id, Bounds::from(coord), value)
+            .await
     }
 }
 
@@ -424,25 +400,22 @@ where
         bounds: Bounds,
         other: DenseTensor<FD, FS, D, T, O>,
     ) -> TCResult<()> {
-        debug!("write dense tensor to dense {}", bounds);
+        debug!("write {} to dense {}", other, bounds);
+        if bounds == Bounds::all(self.shape()) {
+            self.blocks.write(txn, other.blocks).await
+        } else {
+            let slice = self.slice(bounds)?;
+            if other.shape() != slice.shape() {
+                return Err(TCError::unsupported(format!(
+                    "cannot write a tensor of shape {} to a slice of shape {}",
+                    other.shape(),
+                    slice.shape()
+                )));
+            }
 
-        let dtype = self.dtype();
-        let txn_id = *txn.id();
-        let coords = bounds.affected();
-        let slice = self.slice(bounds.clone())?;
-        let other = other.broadcast(slice.shape().clone())?.cast_into(dtype)?;
-
-        let other_values = other.blocks.value_stream(txn).await?;
-        let values = futures::stream::iter(coords).zip(other_values);
-
-        values
-            .map(|(coord, r)| r.map(|value| (coord, value)))
-            .map_ok(|(coord, value)| slice.write_value_at(txn_id, coord, value))
-            .try_buffer_unordered(num_cpus::get())
-            .try_fold((), |_, _| future::ready(Ok(())))
-            .await?;
-
-        Ok(())
+            let bounds = Bounds::all(slice.shape());
+            slice.write(txn, bounds, other).await
+        }
     }
 }
 
@@ -459,6 +432,12 @@ where
     type Txn = T;
 
     async fn mask(self, txn: T, other: Tensor<FD, FS, D, T>) -> TCResult<()> {
+        let other = if self.shape() == other.shape() {
+            other
+        } else {
+            other.broadcast(self.shape().clone())?
+        };
+
         match other {
             Tensor::Dense(dense) => self.mask(txn, dense).await,
             Tensor::Sparse(sparse) => self.mask(txn, sparse.into_dense()).await,
@@ -468,6 +447,13 @@ where
     async fn write(self, txn: T, bounds: Bounds, other: Tensor<FD, FS, D, T>) -> TCResult<()> {
         debug!("DenseTensor::write {} to {}", other, bounds);
 
+        let shape = bounds.to_shape(self.shape())?;
+        let other = if other.shape() == &shape {
+            other
+        } else {
+            other.broadcast(shape)?
+        };
+
         match other {
             Tensor::Dense(dense) => self.write(txn, bounds, dense).await,
             Tensor::Sparse(sparse) => self.write(txn, bounds, sparse.into_dense()).await,
@@ -490,6 +476,7 @@ where
 
     fn add(self, other: DenseTensor<FD, FS, D, T, O>) -> TCResult<Self::Combine> {
         fn add_array(l: &Array, r: &Array) -> Array {
+            debug_assert_eq!(l.len(), r.len());
             l + r
         }
 
@@ -499,6 +486,7 @@ where
 
     fn div(self, other: DenseTensor<FD, FS, D, T, O>) -> TCResult<Self::Combine> {
         fn div_array(l: &Array, r: &Array) -> Array {
+            debug_assert_eq!(l.len(), r.len());
             l / r
         }
 
@@ -508,6 +496,7 @@ where
 
     fn mul(self, other: DenseTensor<FD, FS, D, T, O>) -> TCResult<Self::Combine> {
         fn mul_array(l: &Array, r: &Array) -> Array {
+            debug_assert_eq!(l.len(), r.len());
             l * r
         }
 
@@ -516,7 +505,10 @@ where
     }
 
     fn sub(self, other: DenseTensor<FD, FS, D, T, O>) -> TCResult<Self::Combine> {
+        debug!("subtract {} from {}", other, self);
+
         fn sub_array(l: &Array, r: &Array) -> Array {
+            debug_assert_eq!(l.len(), r.len());
             l - r
         }
 
@@ -553,7 +545,7 @@ where
     fn mul(self, other: Tensor<FD, FS, D, T>) -> TCResult<Self::Combine> {
         match other {
             Tensor::Dense(dense) => self.mul(dense).map(Tensor::from),
-            Tensor::Sparse(sparse) => self.into_sparse().mul(sparse).map(Tensor::from),
+            Tensor::Sparse(sparse) => sparse.mul(self.into_sparse()).map(Tensor::from),
         }
     }
 
@@ -860,9 +852,16 @@ where
     }
 }
 
-impl<FD, FS, D, T, B> fmt::Display for DenseTensor<FD, FS, D, T, B> {
+impl<FD, FS, D, T, B> fmt::Display for DenseTensor<FD, FS, D, T, B>
+where
+    FD: File<Array>,
+    FS: File<Node>,
+    D: Dir,
+    T: Transaction<D>,
+    B: DenseAccess<FD, FS, D, T>,
+{
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        f.write_str("a dense Tensor")
+        write!(f, "a dense Tensor with shape {}", self.shape())
     }
 }
 
@@ -967,13 +966,15 @@ impl<'en> en::IntoStream<'en> for BlockStreamView<'en> {
             ComplexType as CT, FloatType as FT, IntType as IT, NumberType as NT, UIntType as UT,
         };
 
-        fn encodable<'en, E: en::Error + 'en, T: af::HasAfEnum + Clone + Default + 'en>(
+        fn encodable<'en, T: af::HasAfEnum + Clone + Default + 'en>(
             blocks: TCTryStream<'en, Array>,
-        ) -> impl Stream<Item = Result<Vec<T>, E>> + 'en {
+        ) -> impl Stream<Item = Vec<T>> + 'en {
+            // an error can't be encoded within an array
+            // so in case of a read error, let the receiver figure out that the tensor
+            // doesn't have enough elements
             blocks
-                .map_ok(|arr| arr.type_cast())
-                .map_ok(|arr| arr.to_vec())
-                .map_err(en::Error::custom)
+                .take_while(|r| future::ready(r.is_ok()))
+                .map(|r| r.expect("tensor block").type_cast().to_vec())
         }
 
         match self.dtype {
@@ -1005,11 +1006,11 @@ impl<'en> en::IntoStream<'en> for BlockStreamView<'en> {
     }
 }
 
-fn encodable_c32<'en, E: en::Error + 'en>(
-    blocks: TCTryStream<'en, Array>,
-) -> impl Stream<Item = Result<Vec<f32>, E>> + 'en {
+fn encodable_c32<'en>(blocks: TCTryStream<'en, Array>) -> impl Stream<Item = Vec<f32>> + 'en {
     blocks
-        .map_ok(|arr| {
+        .take_while(|r| future::ready(r.is_ok()))
+        .map(|block| block.expect("tensor block"))
+        .map(|arr| {
             let source = arr.type_cast::<afarray::Complex<f32>>();
             let re = source.re();
             let im = source.im();
@@ -1024,14 +1025,13 @@ fn encodable_c32<'en, E: en::Error + 'en>(
 
             dest
         })
-        .map_err(en::Error::custom)
 }
 
-fn encodable_c64<'en, E: en::Error + 'en>(
-    blocks: TCTryStream<'en, Array>,
-) -> impl Stream<Item = Result<Vec<f64>, E>> + 'en {
+fn encodable_c64<'en>(blocks: TCTryStream<'en, Array>) -> impl Stream<Item = Vec<f64>> + 'en {
     blocks
-        .map_ok(|arr| {
+        .take_while(|r| future::ready(r.is_ok()))
+        .map(|block| block.expect("tensor block"))
+        .map(|arr| {
             let source = arr.type_cast::<afarray::Complex<f64>>();
             let re = source.re();
             let im = source.im();
@@ -1046,5 +1046,4 @@ fn encodable_c64<'en, E: en::Error + 'en>(
 
             dest
         })
-        .map_err(en::Error::custom)
 }
