@@ -34,8 +34,8 @@ mod transform;
 
 const PREFIX: PathLabel = path_label(&["state", "collection", "tensor"]);
 
-const ERR_NONBIJECTIVE_WRITE: &str = "Cannot write to a derived Tensor which is not a \
-bijection of its source. Consider copying first, or writing directly to the source Tensor.";
+const ERR_NONBIJECTIVE_WRITE: &str = "cannot write to a derived Tensor which is not a \
+bijection of its source--consider copying first, or writing directly to the source Tensor";
 
 /// The file extension of a [`Tensor`]
 pub const EXT: &str = "array";
@@ -154,11 +154,7 @@ pub trait TensorBoolean<O> {
 }
 
 /// Tensor comparison operations
-#[async_trait]
-pub trait TensorCompare<D: Dir, O> {
-    /// The type of [`Transaction`] to expect
-    type Txn: Transaction<D>;
-
+pub trait TensorCompare<O> {
     /// The result of a comparison operation
     type Compare: TensorInstance;
 
@@ -166,19 +162,19 @@ pub trait TensorCompare<D: Dir, O> {
     type Dense: TensorInstance;
 
     /// Element-wise equality
-    async fn eq(self, other: O, txn: Self::Txn) -> TCResult<Self::Dense>;
+    fn eq(self, other: O) -> TCResult<Self::Dense>;
 
     /// Element-wise greater-than
     fn gt(self, other: O) -> TCResult<Self::Compare>;
 
     /// Element-wise greater-or-equal
-    async fn gte(self, other: O, txn: Self::Txn) -> TCResult<Self::Dense>;
+    fn gte(self, other: O) -> TCResult<Self::Dense>;
 
     /// Element-wise less-than
     fn lt(self, other: O) -> TCResult<Self::Compare>;
 
     /// Element-wise less-or-equal
-    async fn lte(self, other: O, txn: Self::Txn) -> TCResult<Self::Dense>;
+    fn lte(self, other: O) -> TCResult<Self::Dense>;
 
     /// Element-wise not-equal
     fn ne(self, other: O) -> TCResult<Self::Compare>;
@@ -457,8 +453,7 @@ where
     }
 }
 
-#[async_trait]
-impl<FD, FS, D, T> TensorCompare<D, Self> for Tensor<FD, FS, D, T>
+impl<FD, FS, D, T> TensorCompare<Self> for Tensor<FD, FS, D, T>
 where
     FD: File<Array> + TryFrom<D::File, Error = TCError>,
     FS: File<Node> + TryFrom<D::File, Error = TCError>,
@@ -466,14 +461,13 @@ where
     T: Transaction<D>,
     D::FileClass: From<TensorType>,
 {
-    type Txn = T;
     type Compare = Self;
     type Dense = Self;
 
-    async fn eq(self, other: Self, txn: Self::Txn) -> TCResult<Self> {
+    fn eq(self, other: Self) -> TCResult<Self> {
         match self {
-            Self::Dense(dense) => dense.eq(other, txn).await,
-            Self::Sparse(sparse) => sparse.eq(other, txn).await,
+            Self::Dense(dense) => dense.eq(other),
+            Self::Sparse(sparse) => sparse.eq(other),
         }
     }
 
@@ -484,10 +478,10 @@ where
         }
     }
 
-    async fn gte(self, other: Self, txn: Self::Txn) -> TCResult<Self> {
+    fn gte(self, other: Self) -> TCResult<Self> {
         match self {
-            Self::Dense(dense) => dense.gte(other, txn).await,
-            Self::Sparse(sparse) => sparse.gte(other, txn).await,
+            Self::Dense(dense) => dense.gte(other),
+            Self::Sparse(sparse) => sparse.gte(other),
         }
     }
 
@@ -498,10 +492,10 @@ where
         }
     }
 
-    async fn lte(self, other: Self, txn: Self::Txn) -> TCResult<Self> {
+    fn lte(self, other: Self) -> TCResult<Self> {
         match self {
-            Self::Dense(dense) => dense.lte(other, txn).await,
-            Self::Sparse(sparse) => sparse.lte(other, txn).await,
+            Self::Dense(dense) => dense.lte(other),
+            Self::Sparse(sparse) => sparse.lte(other),
         }
     }
 
@@ -917,20 +911,16 @@ impl<FD, FS, D, T> fmt::Display for Tensor<FD, FS, D, T> {
 ///
 /// For rules of broadcasting, see:
 /// [https://pytorch.org/docs/stable/notes/broadcasting.html](https://pytorch.org/docs/stable/notes/broadcasting.html)
-pub fn broadcast<FD, FS, D, T>(
-    left: Tensor<FD, FS, D, T>,
-    right: Tensor<FD, FS, D, T>,
-) -> TCResult<(Tensor<FD, FS, D, T>, Tensor<FD, FS, D, T>)>
+pub fn broadcast<L, R>(left: L, right: R) -> TCResult<(L::Broadcast, R::Broadcast)>
 where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
-    D: Dir,
-    T: Transaction<D>,
-    D::FileClass: From<TensorType>,
+    L: TensorAccess + TensorTransform,
+    R: TensorAccess + TensorTransform,
 {
-    if left.shape() == right.shape() {
-        return Ok((left, right));
-    }
+    debug!(
+        "broadcast tensors with shapes {}, {}",
+        left.shape(),
+        right.shape()
+    );
 
     let mut left_shape = left.shape().to_vec();
     let mut right_shape = right.shape().to_vec();
@@ -963,14 +953,8 @@ where
         }
     }
 
-    let left_shape = Shape::from(left_shape);
-    let right_shape = Shape::from(right_shape);
-    let shape = if left_shape.size() > right_shape.size() {
-        left_shape
-    } else {
-        right_shape
-    };
-
+    let shape = Shape::from(shape);
+    debug!("broadcast shape is {}", shape);
     Ok((left.broadcast(shape.clone())?, right.broadcast(shape)?))
 }
 
@@ -993,3 +977,10 @@ impl<FD, FS, D, T> Default for Phantom<FD, FS, D, T> {
         }
     }
 }
+
+#[inline]
+fn coord_bounds(shape: &[u64]) -> Vec<u64> {
+    (0..shape.len())
+        .map(|axis| shape[axis + 1..].iter().product())
+        .collect()
+}
