@@ -18,7 +18,7 @@ use tc_transact::{IntoView, Transact, Transaction, TxnId};
 use tc_value::{Number, NumberClass, NumberInstance, NumberType};
 use tcgeneric::{TCBoxTryFuture, TCTryStream};
 
-use super::dense::{BlockListFile, BlockListSparse, DenseAccess, DenseTensor};
+use super::dense::{BlockListSparse, DenseTensor};
 use super::{
     Bounds, Coord, Phantom, Schema, Shape, Tensor, TensorAccess, TensorBoolean, TensorCompare,
     TensorDualIO, TensorIO, TensorInstance, TensorMath, TensorReduce, TensorTransform, TensorType,
@@ -35,7 +35,6 @@ mod access;
 mod combine;
 mod table;
 
-type CoordStream<'a> = Pin<Box<dyn Stream<Item = TCResult<Coord>> + Send + Unpin + 'a>>;
 pub type SparseRow = (Coord, Number);
 pub type SparseStream<'a> = Pin<Box<dyn Stream<Item = TCResult<SparseRow>> + Send + Unpin + 'a>>;
 
@@ -56,6 +55,9 @@ impl<FD, FS, D, T, A> SparseTensor<FD, FS, D, T, A> {
     }
 }
 
+type Condensed<FD, FS, D, T, L, R> =
+    DenseTensor<FD, FS, D, T, BlockListSparse<FD, FS, D, T, SparseCombinator<FD, FS, D, T, L, R>>>;
+
 impl<FD, FS, D, T, A> SparseTensor<FD, FS, D, T, A>
 where
     FD: File<Array>,
@@ -72,7 +74,7 @@ where
     ) -> TCResult<SparseTensor<FD, FS, D, T, SparseCombinator<FD, FS, D, T, A, R>>> {
         if self.shape() != other.shape() {
             return Err(TCError::unsupported(format!(
-                "cannot combine Tensors of different shapes: {}, {}",
+                "cannot compare Tensors of different shapes: {}, {}",
                 self.shape(),
                 other.shape()
             )));
@@ -85,56 +87,28 @@ where
             phantom: self.phantom,
         })
     }
-}
 
-impl<FD, FS, D, T, A> SparseTensor<FD, FS, D, T, A>
-where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
-    D: Dir,
-    T: Transaction<D>,
-    A: SparseAccess<FD, FS, D, T>,
-    D::FileClass: From<TensorType>,
-{
-    fn condense<'a, R>(
+    fn condense<R>(
         self,
         other: SparseTensor<FD, FS, D, T, R>,
-        txn: T,
         condensor: fn(Number, Number) -> Number,
-    ) -> TCBoxTryFuture<'a, DenseTensor<FD, FS, D, T, BlockListFile<FD, FS, D, T>>>
+    ) -> TCResult<Condensed<FD, FS, D, T, A, R>>
     where
         R: SparseAccess<FD, FS, D, T>,
     {
-        Box::pin(async move {
-            if self.shape() != other.shape() {
-                return Err(TCError::unsupported(format!(
-                    "cannot condense sparse Tensor of size {} with another of size {}",
-                    self.shape(),
-                    other.shape()
-                )));
-            }
+        if self.shape() != other.shape() {
+            return Err(TCError::unsupported(format!(
+                "cannot condense sparse Tensor of size {} with another of size {}",
+                self.shape(),
+                other.shape()
+            )));
+        }
 
-            let shape = self.shape().clone();
-            let accessor =
-                SparseCombinator::new(self.accessor, other.accessor, condensor, NumberType::Bool)?;
+        let accessor =
+            SparseCombinator::new(self.accessor, other.accessor, condensor, NumberType::Bool)?;
 
-            let txn_id = *txn.id();
-            let file = txn
-                .context()
-                .create_file_tmp(txn_id, TensorType::Dense)
-                .await?;
-
-            let condensed = DenseTensor::constant(file, txn_id, shape, true.into()).await?;
-
-            let filled = accessor.filled_inner(txn).await?;
-            filled
-                .map_ok(|(coord, value)| condensed.write_value_at(txn_id, coord, value))
-                .try_buffer_unordered(num_cpus::get())
-                .try_fold((), |_, _| future::ready(Ok(())))
-                .await?;
-
-            Ok(condensed)
-        })
+        let dense = BlockListSparse::from(accessor);
+        Ok(dense.into())
     }
 }
 
@@ -252,8 +226,7 @@ where
     }
 }
 
-#[async_trait]
-impl<FD, FS, D, T, L, R> TensorCompare<D, SparseTensor<FD, FS, D, T, R>>
+impl<FD, FS, D, T, L, R> TensorCompare<SparseTensor<FD, FS, D, T, R>>
     for SparseTensor<FD, FS, D, T, L>
 where
     FD: File<Array> + TryFrom<D::File, Error = TCError>,
@@ -264,20 +237,15 @@ where
     R: SparseAccess<FD, FS, D, T>,
     D::FileClass: From<TensorType>,
 {
-    type Txn = T;
     type Compare = SparseTensor<FD, FS, D, T, SparseCombinator<FD, FS, D, T, L, R>>;
-    type Dense = DenseTensor<FD, FS, D, T, BlockListFile<FD, FS, D, T>>;
+    type Dense = Condensed<FD, FS, D, T, L, R>;
 
-    async fn eq(
-        self,
-        other: SparseTensor<FD, FS, D, T, R>,
-        txn: Self::Txn,
-    ) -> TCResult<Self::Dense> {
+    fn eq(self, other: SparseTensor<FD, FS, D, T, R>) -> TCResult<Self::Dense> {
         fn eq(l: Number, r: Number) -> Number {
             (l == r).into()
         }
 
-        self.condense(other, txn, eq).await
+        self.condense(other, eq)
     }
 
     fn gt(self, other: SparseTensor<FD, FS, D, T, R>) -> TCResult<Self::Compare> {
@@ -288,16 +256,12 @@ where
         self.combine(other, gt, NumberType::Bool)
     }
 
-    async fn gte(
-        self,
-        other: SparseTensor<FD, FS, D, T, R>,
-        txn: Self::Txn,
-    ) -> TCResult<Self::Dense> {
+    fn gte(self, other: SparseTensor<FD, FS, D, T, R>) -> TCResult<Self::Dense> {
         fn gte(l: Number, r: Number) -> Number {
             (l >= r).into()
         }
 
-        self.condense(other, txn, gte).await
+        self.condense(other, gte)
     }
 
     fn lt(self, other: SparseTensor<FD, FS, D, T, R>) -> TCResult<Self::Compare> {
@@ -308,16 +272,12 @@ where
         self.combine(other, lt, NumberType::Bool)
     }
 
-    async fn lte(
-        self,
-        other: SparseTensor<FD, FS, D, T, R>,
-        txn: Self::Txn,
-    ) -> TCResult<Self::Dense> {
+    fn lte(self, other: SparseTensor<FD, FS, D, T, R>) -> TCResult<Self::Dense> {
         fn lte(l: Number, r: Number) -> Number {
             (l <= r).into()
         }
 
-        self.condense(other, txn, lte).await
+        self.condense(other, lte)
     }
 
     fn ne(self, other: SparseTensor<FD, FS, D, T, R>) -> TCResult<Self::Compare> {
@@ -329,8 +289,7 @@ where
     }
 }
 
-#[async_trait]
-impl<FD, FS, D, T, A> TensorCompare<D, Tensor<FD, FS, D, T>> for SparseTensor<FD, FS, D, T, A>
+impl<FD, FS, D, T, A> TensorCompare<Tensor<FD, FS, D, T>> for SparseTensor<FD, FS, D, T, A>
 where
     FD: File<Array> + TryFrom<D::File, Error = TCError>,
     FS: File<Node> + TryFrom<D::File, Error = TCError>,
@@ -339,14 +298,13 @@ where
     A: SparseAccess<FD, FS, D, T>,
     D::FileClass: From<TensorType>,
 {
-    type Txn = T;
     type Compare = Tensor<FD, FS, D, T>;
     type Dense = Tensor<FD, FS, D, T>;
 
-    async fn eq(self, other: Tensor<FD, FS, D, T>, txn: Self::Txn) -> TCResult<Self::Dense> {
+    fn eq(self, other: Tensor<FD, FS, D, T>) -> TCResult<Self::Dense> {
         match other {
-            Tensor::Dense(other) => self.eq(other.into_sparse(), txn).map_ok(Tensor::from).await,
-            Tensor::Sparse(other) => self.eq(other, txn).map_ok(Tensor::from).await,
+            Tensor::Dense(other) => self.eq(other.into_sparse()).map(Tensor::from),
+            Tensor::Sparse(other) => self.eq(other).map(Tensor::from),
         }
     }
 
@@ -357,10 +315,10 @@ where
         }
     }
 
-    async fn gte(self, other: Tensor<FD, FS, D, T>, txn: Self::Txn) -> TCResult<Self::Dense> {
+    fn gte(self, other: Tensor<FD, FS, D, T>) -> TCResult<Self::Dense> {
         match other {
-            Tensor::Dense(other) => self.into_dense().gte(other, txn).map_ok(Tensor::from).await,
-            Tensor::Sparse(other) => self.gte(other, txn).map_ok(Tensor::from).await,
+            Tensor::Dense(other) => self.into_dense().gte(other).map(Tensor::from),
+            Tensor::Sparse(other) => self.gte(other).map(Tensor::from),
         }
     }
 
@@ -371,10 +329,10 @@ where
         }
     }
 
-    async fn lte(self, other: Tensor<FD, FS, D, T>, txn: Self::Txn) -> TCResult<Self::Dense> {
+    fn lte(self, other: Tensor<FD, FS, D, T>) -> TCResult<Self::Dense> {
         match other {
-            Tensor::Dense(other) => self.into_dense().lte(other, txn).map_ok(Tensor::from).await,
-            Tensor::Sparse(other) => self.lte(other, txn).map_ok(Tensor::from).await,
+            Tensor::Dense(other) => self.into_dense().lte(other).map(Tensor::from),
+            Tensor::Sparse(other) => self.lte(other).map(Tensor::from),
         }
     }
 
@@ -386,36 +344,6 @@ where
     }
 }
 
-#[async_trait]
-impl<FD, FS, D, T, A, B> TensorDualIO<D, DenseTensor<FD, FS, D, T, B>>
-    for SparseTensor<FD, FS, D, T, A>
-where
-    FD: File<Array> + TryFrom<D::File, Error = TCError>,
-    FS: File<Node> + TryFrom<D::File, Error = TCError>,
-    D: Dir,
-    T: Transaction<D>,
-    A: SparseAccess<FD, FS, D, T>,
-    B: DenseAccess<FD, FS, D, T>,
-    D::FileClass: From<TensorType>,
-{
-    type Txn = T;
-
-    async fn mask(self, txn: Self::Txn, other: DenseTensor<FD, FS, D, T, B>) -> TCResult<()> {
-        let other = other.into_sparse();
-        self.mask(txn, other).await
-    }
-
-    async fn write(
-        self,
-        txn: Self::Txn,
-        bounds: Bounds,
-        other: DenseTensor<FD, FS, D, T, B>,
-    ) -> TCResult<()> {
-        let other = other.into_sparse();
-        self.write(txn, bounds, other).await
-    }
-}
-
 #[async_trait]
 impl<FD, FS, D, T, L, R> TensorDualIO<D, SparseTensor<FD, FS, D, T, R>>
     for SparseTensor<FD, FS, D, T, L>
@@ -460,7 +388,7 @@ where
         let slice = self.slice(bounds)?;
         if slice.shape() != other.shape() {
             return Err(TCError::unsupported(format!(
-                "cannot write Tensor with shape {} to slice with shape {}",
+                "cannot write tensor of shape {} to slice of shape {}",
                 other.shape(),
                 slice.shape()
             )));
@@ -489,8 +417,14 @@ where
     type Txn = T;
 
     async fn mask(self, txn: Self::Txn, other: Tensor<FD, FS, D, T>) -> TCResult<()> {
+        let other = if self.shape() == other.shape() {
+            other
+        } else {
+            other.broadcast(self.shape().clone())?
+        };
+
         match other {
-            Tensor::Dense(other) => self.mask(txn, other).await,
+            Tensor::Dense(other) => self.mask(txn, other.into_sparse()).await,
             Tensor::Sparse(other) => self.mask(txn, other).await,
         }
     }
@@ -501,8 +435,15 @@ where
         bounds: Bounds,
         other: Tensor<FD, FS, D, T>,
     ) -> TCResult<()> {
+        let shape = bounds.to_shape(self.shape())?;
+        let other = if other.shape() == &shape {
+            other
+        } else {
+            other.broadcast(shape)?
+        };
+
         match other {
-            Tensor::Dense(other) => self.write(txn, bounds, other).await,
+            Tensor::Dense(other) => self.write(txn, bounds, other.into_sparse()).await,
             Tensor::Sparse(other) => self.write(txn, bounds, other).await,
         }
     }
@@ -527,6 +468,10 @@ where
     }
 
     async fn write_value(&self, txn_id: TxnId, mut bounds: Bounds, value: Number) -> TCResult<()> {
+        if self.shape().is_empty() {
+            return self.accessor.write_value(txn_id, vec![], value).await;
+        }
+
         bounds.normalize(self.shape());
         debug!("SparseTensor::write_value {} to bounds, {}", value, bounds);
         stream::iter(bounds.affected())
@@ -559,11 +504,13 @@ where
     }
 
     fn div(self, other: SparseTensor<FD, FS, D, T, R>) -> TCResult<Self::Combine> {
+        // TODO: implement left_combine
         let dtype = Ord::max(self.dtype(), other.dtype());
         self.combine(other, Number::div, dtype)
     }
 
     fn mul(self, other: SparseTensor<FD, FS, D, T, R>) -> TCResult<Self::Combine> {
+        // TODO: implement left_combine
         let dtype = Ord::max(self.dtype(), other.dtype());
         self.combine(other, Number::mul, dtype)
     }
